// USB+Ethernet/DMX Generator  ***********************************************
//
// Simple DMX512(incl. Art-Net) controller for STM32 Nucleo-L476 + PCB
// based on USB MIDISTREAM and CDC-ABSTRACT CONTROL MODEL
//
// supports the essential features of Art-Net, the MiniDMX protocol
// plus simple MIDI message and ASCII text based command sets
//
// Author and (c)2017 Wolfgang Schemmert. The copyright claim and GPL-3 Licence
// is restricted to this file and the depending files
// "main.h", "usb-otg.c", "dmx.c", "ether.c" and "userinterface.c".
// Other depending material is published and copyrighted
// by STMicroelectronics, Atollic AB and EMBITZ.org
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual is provided at this website
//
// Note: at present state of art, only the DEBUG object code is working.
// The RELEASE version compiled with EmBitz1.11 does not work. Reason unkown.
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// **************************************************************************

#include "stm32l4xx.h"
#include "globaldefines.h"
#include "main.h"

int main(void)    // ********************************************************
// **************************************************************************
{
uint32_t recvmsg;
uint16_t recvword;
uint8_t recvbyte;


	SCB->CCR |= SCB_CCR_STKALIGN_Msk; // Enable double word stack alignment
//(recommended in Cortex-M3 r1p1, default in Cortex-M3 r2px and Cortex-M4)

// effectively executed code of NVIC_PriorityGroupConfig(uint32_t NVIC_PriorityGroup) :
#define AIRCR_VECTKEY_MASK    ((uint32_t)0x05FA0000)
#define NVIC_PriorityGroup_0  ((uint32_t)0x700) // < 0 bits for pre-emption priority
                                                // 4 bits for subpriority
    // Set the PRIGROUP[10:8] bits according to NVIC_PriorityGroup value
    SCB->AIRCR = AIRCR_VECTKEY_MASK | (uint32_t)0x700;

    SetSysClock(); //special for this CPU,see below at "Misc Low Level Routines"

//init I/O state  *********************
	RCC->AHB2ENR |= RCC_AHB2ENR_GPIOAEN;  // Enable Port A clock
	RCC->AHB2ENR |= RCC_AHB2ENR_GPIOBEN;  // Enable Port B clock
	RCC->AHB2ENR |= RCC_AHB2ENR_GPIOCEN;  // Enable Port C clock
	RCC->AHB2ENR |= RCC_AHB2ENR_GPIODEN;  // Enable Port D clock

//STM32L476 port default mode = ANALOG !!
 	GPIOA->MODER &= ~(GPIO_MODER_MODE15_0) ; //disable JTAG JTDI
    GPIOA->MODER &= ~(GPIO_MODER_MODE15_1) ;
    GPIOB->MODER = 0;               // first set all pins to input, disable JTAG
    GPIOC->MODER = 0;                     // first set all pins to input
    //GPIOx->OTYPER Pins are push-pull by default
    //GPIOx->OSPEEDR set to 2 MHz by default
    //GPIOC->PUPDR Pins are HiZ by default
    GPIOC->MODER |= GPIO_MODER_MODE4_0;     //Set PC4 = Reset pin as output
    GPIOC->BSRR =  GPIO_BSRR_BS4;           //Reset pin High
    GPIOC->MODER |= GPIO_MODER_MODE8_0;  //Set PC8 = optional test pin as output
    GPIOC->BSRR =  GPIO_BSRR_BS8;           //test pin High
    GPIOC->MODER |= GPIO_MODER_MODE2_0 ;    //Set PortC2 (green LED) as output
    GPIOC->BSRR =  GPIO_BSRR_BS2;           //LED ON
    GPIOC->MODER |= GPIO_MODER_MODE3_0 ;    //Set PortC3 (red LED) as output
    GPIOC->BSRR =  GPIO_BSRR_BS3;           //LED ON

// PortC,7 enables/disables the DMX driver
    GPIOC->MODER |= GPIO_MODER_MODER7_0 ; //PortC7 general purpose output mode
    GPIOC->BSRR = GPIO_BSRR_BR_7;         //clear bit PC7 (DMX driver disabled)
    //GPIOx->OTYPER  is push-pull by default
    //GPIOx->OSPEEDR = 0;  //2 MHz by default
    //GPIOx->PUPDR  is HiZ by default

    SysTick_Config(SystemCoreClock/1000);

    gdMiniDmxOn = 0;
    gcIgnoreVelo0 = 0;
    gVidShadow = 0;             //as precaution, will be set finally
    gPidShadow = 0;             //by popFlashPage15()
    gSYSCOUNT=20;
    gGREEN_LEDCOUNT=0;
    gRED_LEDCOUNT=0;
    gDELAYCOUNT=0;
    gTIMEOUT=0;
    gdDmxMax = 512;
    gdFadeTime = 0;

#ifndef DEBUG
//start Independent Watchdog  **************
    IWDG->KR = 0x00005555;
    IWDG->RLR = 0x0000FFFF;
    msDelay(2);
    IWDG->PR = 0x00000002;      //resulting timeout ca. 1-2 seconds
    msDelay(2);
    IWDG->KR = 0x0000CCCC;      //start Independent Watchdog
#endif

//check DIPSW and set mode of operation ****************************
//PortC is already configured as input, i.e. MODEx_0 and MODEx_1=0
//configure PortC 0,1 as pull-up
    GPIOC->PUPDR |=  (GPIO_PUPDR_PUPD0_0 | GPIO_PUPDR_PUPD1_0);
    GPIOC->PUPDR &= ~(GPIO_PUPDR_PUPD0_1 | GPIO_PUPDR_PUPD1_1);
    msDelay(2);

    gOpMode = ~(GPIOC->IDR & (GPIO_IDR_IDR_0 | GPIO_IDR_IDR_1)) & 3;
//gOpMode=0: both switches open = default Config mode-USB or USART2
//gOpMode=1:left switch closed, =2:right switch closed, =3:both closed
//gOpMode=0: ASCII setup via USB, provisionally in parallel via USART2
//gOpMode=1: Artnet and parallel ASCII commands, Mini DMX via USB
//gOpMode=2: only Artnet
//gOpMode=3: Artnet and parallel MIDI commands via USB

    guDeviceClass = CDC_CLASS;
    geArtnetOn = 1;                         //default, my be blocked manually
    if (gOpMode == 3)  {
        guDeviceClass = MIDI_CLASS;
        geArtnetOn = 0;                     //may be unblocked manually
    }

//init communication  *********************
    initUsart2();               //parallel with USB, especially for debugging
    msDelay(2);

#ifdef DEBUG
    gDBGHEAD=gDBGBUF;           //init buffer for sdbgX..() debug messages
#endif
    popFlashPage15();

    guConfigured = 0;
    guUsbCmdEnable = 0;

    initUSB();
    msDelay(50);

    initSPI1();
    msDelay(100);
    initW5500();
    msDelay(100);
    openUDP0();

    gTIMEOUT = 5000;                    //wait for max 5 seconds
    while (gTIMEOUT)  {
        IWDG->KR = 0x0000AAAA;          //feed watchdog
        if (guConfigured)  break;
    }

    if ((gOpMode == 0) && (guConfigured == 0))  sendString(gNoUsbStr);

    GPIOC->BSRR = GPIO_BSRR_BS_7;       //enable DMX driver
    startDmxTx();                       //transmits even in Setup Mode
    DmxHeader();

    gcPushByte = 0;

    while(1)  {
        servSystem();                   //essential, controls DMX timing

        if (gOpMode <= 1)  {
            if (gcPushByte)  {
                recvbyte = gcPushByte;
                gcPushByte = 0;
                if (gOpMode == 0) Setup_Interpreter(recvbyte);
                else AsciiCmd_Interpreter(recvbyte);
            }
            recvword = pullCmdRxBuf();
            if (recvword)  {
                recvbyte = (uint8_t)(recvword & 0x00FF);
                if ((gOpMode==1) && (recvbyte == 'Z'))  {
                    geArtnetOn = 0;                 //block always
                    gdFaderBusy = 0;                //block fade process
                    gTIMEOUT = 1000;
                    MiniDMX_Interpreter();
                }
                else if (gOpMode == 0) Setup_Interpreter(recvbyte);
                else AsciiCmd_Interpreter(recvbyte);
            }
            else if ((gTIMEOUT == 0) && gdMiniDmxOn) {
                gdMiniDmxOn = 0;    //terminate MiniDMX
                geArtnetOn = 1;     //re-activate Art-Net
            }
        }
        else if (gOpMode == 3) {
            recvmsg = pullMidiMsg() ;
            if (recvmsg) MidiCmd_Interpreter(recvmsg);
        }
        //no manual control if gOpMode = 2,-->Artnet only

#ifdef DEBUG
            if ((recvbyte&0x00FF)==0x7E)  sendDebug();
//Tilde char ~ transmits and flushes the debug buffer.
//It transmits and flushes automatically when it is full.
#endif

    }
}


//##########################################################################

// Misc Low Level Routines

// ************************

#ifdef MSI_CLOCK_DIRECT   //  !!!!!!!!  #####################################

void SetSysClock(void)  // **************************************************
//this method is the only one which works in practice (at least together with USB)
//provides "Vcore range 1", MSI clock 48.000 MHz, no PLL, AHB=SystemClock, APB1+2 = 24 MHz
// **************************************************************************
{

//enable power to the backup domain (LSE oscillator)
    RCC->APB1ENR1 |= RCC_APB1ENR1_PWREN;
//adjust Vcore Range1 and disable backup domain write protection
    while (PWR->SR2 & PWR_SR2_VOSF) ;
    PWR->CR1 &= ~PWR_CR1_VOS_1;
    PWR->CR1 |= (PWR_CR1_VOS_0 | PWR_CR1_DBP);
    while (PWR->SR2 & PWR_SR2_VOSF) ;

//adjust Flash Waitstates and Prefetch for higher System Clock frequency
    FLASH->ACR |= (FLASH_ACR_PRFTEN | FLASH_ACR_LATENCY_2WS); //for SYSCLK=48MHz

    RCC->CR &= ~RCC_CR_PLLON;           //stop PLL, MSI keeps running with default 4MHz
    while(RCC->CR & RCC_CR_PLLRDY) ;    //as a measure of precaution

    RCC->BDCR &= ~0x0000003D;           //reset all LSE relevant bits
//start LSE ocscillator
    RCC->BDCR |= (RCC_BDCR_LSEDRV_0 | RCC_BDCR_LSEDRV_1);     //high drive
    RCC->BDCR |= RCC_BDCR_LSEON;        //start
    while (!(RCC->BDCR & RCC_BDCR_LSERDY));

//adjust MSI oscillator
    while (!(RCC->CR & RCC_CR_MSIRDY)); //as precaution, should be ok
    RCC->CR &= ~RCC_CR_MSIRANGE;        //clear all relevant bits
    RCC->CR |= RCC_CR_MSIRANGE_11;      //set new value
    RCC->CR |= (RCC_CR_MSIRGSEL | RCC_CR_MSIPLLEN | RCC_CR_MSION);
    while (!(RCC->CR & RCC_CR_MSIRDY)); //confirm PLL lockin at new range

//SYSCLK=MSI and HCLK = SYSCLK, already set in "SystemInit()"
    RCC->CFGR |= (RCC_CFGR_PPRE1_DIV2 | RCC_CFGR_PPRE2_DIV2);
// PCLK1 = HCLK/2 = 24MHz
// PCLK2 = HCLK/2 = 24MHz

// select MSI as clock for USB
    RCC->CCIPR |= RCC_CCIPR_CLK48SEL;    //set CLK48SEL[1:0]=11

    SystemCoreClock = 48000000;

}

#endif  //MSI_CLOCK_DIRECT


void msDelay(uint16_t count)   // *******************************************
//blocking delay by 'count' milliseconds
// **************************************************************************
{
    gDELAYCOUNT = count;
    while(gDELAYCOUNT) IWDG->KR = 0x0000AAAA;
                                    //feed watchdog, timeout ca. 1-2 seconds;
}


void wait1us(void)   // *****************************************************
//blocking delay by a little bit more than 1 microsecond
// **************************************************************************
{
uint32_t countdown;


	countdown = (uint32_t)SystemCoreClock/12000000;
    while (countdown)  countdown-- ;
}


void servSystem(void)   // ************************************************
//not very time critical but permanently repeated functions for system control
//exclusively called in pullMIDIxRxBuf, and pullUsbRxBuf
//this way it is regularily served when idle waiting for external input
//but watchdog resets when it hangs somewhere else
//additionally slowed down by SysTick to 50 times/second
// ****************************************************************************
{
uint8_t temp;
static uint8_t secCount;
uint16_t switches;

//system relevant, but not extremely time critical services
//are handled quasi in the background

    receiveUDP();   //Art-Net is the only UDP feature

    if (USART1->ISR & USART_ISR_TXE) {
        USART1->TDR = gdDMXBUF[gdDmxTxIndex];
// the TXE bit is cleared automatically when TDR is written
        gdDmxTxIndex++;
        if (gdDmxTxIndex >= (gdDmxMax + 1))  DmxHeader() ; // >= new 30July20
    }

    if (gSYSCOUNT == 0)  {
//lower speed services:
        gSYSCOUNT = 20;         //restart 50Hz service interval
//feed watchdog
        IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds

//handle DMX faders
        if (gdFaderBusy)  fadeProcess();

//LED control
        if (gGREEN_LEDCOUNT==0)  GPIOC->BSRR =  GPIO_BSRR_BS_2;
                                                    //PC2 (red LED) ON again
        if (gRED_LEDCOUNT==0)  GPIOC->BSRR =  GPIO_BSRR_BS_3;
                                                    //PC3 (green LED) ON again
        secCount ++;
        if (secCount >= 50)  {
//very low speed service (1 per sec)
            secCount = 0;

#ifndef DEBUG

//check DIP switch and force new USB enumeration if necessary
//gOpMode=0: both switches open = default Config mode-USB or USART2
//gOpMode=1:left switch closed, =2:right switch closed, =3:both closed
//gOpMode=0: ASCII setup via USB, provisionally in parallel via USART2
//gOpMode=1: Artnet and parallel  ASCII commands, Mini DMX via USB
//gOpMode=2: only Artnet
//gOpMode=3: Artnet and parallel MIDI commands via USB
            switches= ~(GPIOC->IDR & (GPIO_IDR_IDR_0 | GPIO_IDR_IDR_1)) & 3;
            temp = gOpMode;
            if (switches != gOpMode)  {
                gOpMode = switches;
                if(((temp < 3)&&(gOpMode == 3)) ||
                   ((temp == 3)&&(gOpMode < 3))) {
                    GPIOC->BSRR =  GPIO_BSRR_BR_2 | GPIO_BSRR_BR_3; //LEDs OFF
                    GPIOC->BSRR = GPIO_BSRR_BR_7;       //disable DMX driver
                    stopUSB();
                    msDelay(500);
                    if (gOpMode < 3) guDeviceClass = CDC_CLASS;
                    else guDeviceClass = MIDI_CLASS;
                    IWDG->KR = 0x0000AAAA;              //feed watchdog
                    initUSB();
                    GPIOC->BSRR = GPIO_BSRR_BS_7;       //enable DMX driver

                }
                gTIMEOUT = 5000;                    //max 5 seconds
                while (gTIMEOUT)  {
                    IWDG->KR = 0x0000AAAA;          //feed watchdog
                    if (guConfigured)  break;
                }
                geArtnetOn = 1;      //default 1= don't block Artnet, 0=block
                if (gOpMode == 3) geArtnetOn = 0; //may be unblocked manually
                GPIOC->BSRR =  GPIO_BSRR_BS_2 | GPIO_BSRR_BS_3;     //LEDs ON
            }
#endif
        }
    }
}


void popFlashPage15(void)   // **********************************************
//only the first 12 words = 48 bytes of the page are read
// **************************************************************************
{
uint8_t k;
uint32_t *flashPtr, *comParamPtr;

    flashPtr = (uint32_t *)PAGE15;
    comParamPtr = (uint32_t *) &(gComParam.VidPid);
    for (k=0;k<10;k++)  {
        *comParamPtr = ~(*flashPtr);
        comParamPtr++;
        flashPtr++;
    }
//if flash item is empty, initialize default
 //   if (gComParam.VidPid == 0) gComParamShadow.VidPid = VIDPID_DEFAULT;
    if (gComParam.MacHi == 0) gComParam.MacHi = MACHI_DEFAULT;
    if (gComParam.MacLo == 0) gComParam.MacLo = MACLO_DEFAULT;
    if (gComParam.Gateway == 0) gComParam.Gateway = GATEWAY_DEFAULT;
    if (gComParam.LocalIP == 0) gComParam.LocalIP = LOCALIP_DEFAULT;
    if (gComParam.Subnet == 0) gComParam.Subnet = SUBNET_DEFAULT;
 //   if (gComParam.ArtnetIP==0) gComParam.ArtnetIP = 2;
 //   if (gComParam.ArtSubnet==0) gComParam.ArtSubnet=0;
 //   if (gComParam.ArtUniverse==0) gComParam.ArtUniverse=0;
//   if (gComParam.Channel==0) gComParam.Channel=0;
    if (gComParam.VidPid == 0) {
        gVidShadow = VID_DEFAULT;
        gPidShadow = PID_DEFAULT;
    }
    else  {
        gVidShadow = (uint16_t)(gComParam.VidPid >> 16);
        gPidShadow = (uint16_t)(gComParam.VidPid & 0x0000FFFF);
    }
}


void pushFlashPage15(void)   // ********************************************
//only the first  12 words = 48 bytes of the page are written
// **************************************************************************
{
uint8_t k;
uint32_t *flashPtr, *comParamPtr;

//unlock flash
#define KEY1    0x45670123      //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB

    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
                                //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
            FLASH->KEYR = KEY1;
            FLASH->KEYR = KEY2;
    }

//erase Page15
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = ((15<<3) | FLASH_CR_EOPIE |FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) !=0)  FLASH->SR |= FLASH_SR_EOP;
    else {
        sendX32(FLASH->SR);        //error management ?
        sendCmdB('?');
        sendCmdB('?');
    }
    FLASH->CR &= ~FLASH_CR_PER;

    gComParam.VidPid =  ((uint32_t)gVidShadow << 16) | (uint32_t)gPidShadow;

    flashPtr = (uint32_t *)PAGE15;
    comParamPtr = (uint32_t *) &(gComParam.VidPid);
    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
// flash least word is first word in struct
// this is compatible with flash readout in words
    for (k=0;k<5;k++)  {
        *flashPtr = ~(*comParamPtr);
        comParamPtr++;
        flashPtr++;
        *flashPtr = ~(*comParamPtr);
        comParamPtr++;
        flashPtr++;
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        else {
            sendX8(k);
            sendX32(FLASH->SR);        //error management ?
            sendCmdB('!');
            sendCmdB('!');
        }
    }
    FLASH->CR &= ~FLASH_CR_PG;
//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);
}


void popPresetPage(uint32_t presetNum)   // *********************************
//copy  first 512 bytes from selected flash page (presetNum+16)
//into actual DMX transmit buffer
// **************************************************************************
{
uint8_t k;
uint32_t *flashPtr, *dmxPtr;

    flashPtr = (uint32_t *)(PAGE16 + presetNum*0x800);
    dmxPtr = (uint32_t *)gdFINALMEM;
    for (k=0;k<130;k++)  {                          //i.e.copy 516 bytes
        *dmxPtr = ~(*flashPtr++);                   //load one's complement
        dmxPtr++;
    }

    if (presetNum == 0) {
        gdDmxMax =
        (uint16_t)(gdFINALMEM[DMX_BUFLEN+2] << 8) | gdFINALMEM[DMX_BUFLEN+3];
        if (gdDmxMax == 0) gdDmxMax = 512;        //bugfix 30July20
        if (gdDmxMax < 24) gdDmxMax = 24;
        gdFadeTime =
        (uint16_t)(gdFINALMEM[DMX_BUFLEN+4] << 8) | gdFINALMEM[DMX_BUFLEN+5];
        if (gdFadeTime > 127) gdFadeTime = 127;   //new 30July20
    }

    fade2Preset();
}


void pushPresetPage(uint32_t presetNum)   // *********************************
//write actual DMX transmit buffer (512 bytes=64 double words)
//into selected flash page (presetNum+16)
// **************************************************************************
{
uint8_t k;
uint32_t *flashPtr,*dmxBufPtr;

//unlock flash
#define KEY1    0x45670123      //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB

    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
                                //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
            FLASH->KEYR = KEY1;
            FLASH->KEYR = KEY2;
    }

//erase Page (presetNum+16)
    k = presetNum + 16;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = ((k<<3) | FLASH_CR_EOPIE |FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    if ((FLASH->SR & FLASH_SR_EOP) !=0)  FLASH->SR |= FLASH_SR_EOP;
    else {
        sendX32(FLASH->SR);        //error management ?
        sendCmdB('?');
        sendCmdB('?');
    }
    FLASH->CR &= ~FLASH_CR_PER;

//write DMX buffer into Page (presetNum+16)
    gdDMXBUF[DMX_BUFLEN+2] = (gdDmxMax>>8);       //insert gdDmxMax
    gdDMXBUF[DMX_BUFLEN+3] = (gdDmxMax& 0x00FF);
    gdDMXBUF[DMX_BUFLEN+4] = (gdFadeTime>>8);     //insert gdFadeTime
    gdDMXBUF[DMX_BUFLEN+5] = (gdFadeTime& 0x00FF);

    flashPtr = (uint32_t *)(PAGE16 + presetNum*0x800);      //address page
    dmxBufPtr = (uint32_t *)gdDMXBUF;
    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
// flash least word is first word in struct
// this is compatible with flash readout in words
    for (k=0;k<65;k++)  {
        *flashPtr = ~(*dmxBufPtr++);
        flashPtr++;
        *flashPtr = ~(*dmxBufPtr++);
        flashPtr++;
        while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
        if ((FLASH->SR & FLASH_SR_EOP) != 0)  FLASH->SR |= FLASH_SR_EOP;
        else {
            sendX8(k);
            sendX32(FLASH->SR);        //error management ?
            sendCmdB('!');
            sendCmdB('!');
        }
    }
    FLASH->CR &= ~FLASH_CR_PG;
//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);
}

//###########################################################################

//USART Support
//*************

void initUsart2(void) // ****************************************************
// USART2-TxD is PortA,2
// USART2-RxD is PortA,3
// **************************************************************************
{
uint16_t baudset;

    RCC->APB1ENR1 |=  RCC_APB1ENR1_USART2EN ;   // USART2 clock enable

//STM32L476 port default mode = ANALOG !!
	GPIOA->MODER &= ~(GPIO_MODER_MODE2_0 |    // Set pin 2,3 in alternate function mode
	                 GPIO_MODER_MODE3_0 ) ;  // 10=Alternate function mode
	GPIOA->MODER |= (GPIO_MODER_MODE2_1 |    // Set pin 2,3 in alternate function mode
	                 GPIO_MODER_MODE3_1 ) ;  // 10=Alternate function mode
    //GPIOA->OTYPER is push-pull by default
    //GPIOA->OSPEEDR is 2 MHz by default
    //GPIOA->PUPDR is HiZ by default (except PA13,14,15)

//USART2 is AF7
    GPIOA->AFR[0] &= 0xFFFF00FF;
    GPIOA->AFR[0] |= 0x00007700;

    gUSART_RXHEAD = gUSART_RXBUF;
    gUSART_RXTAIL = gUSART_RXBUF;

    baudset = 0x00D0;           //115200Baud: 13.0 at 24 MHz APB1 Clock
    USART2->BRR = baudset;      //only option when used together with onboard ST-Link

    USART2->CR2 = 0;			//default:1 Stopbit
    USART2->CR3 = 0;			//default:kein Handshake
    USART2->CR1 = USART_CR1_UE | USART_CR1_RXNEIE |USART_CR1_TE | USART_CR1_RE;
                                                        //8Bit, no Parity, RX Int
    NVIC_EnableIRQ(USART2_IRQn);
}


void sendB2(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
    while(!(USART2->ISR&USART_ISR_TXE));
    USART2->TDR = data;
}


uint16_t pullUsartRxBuf(void) // ********************************************
// pull 1 byte asynchronously from USART reveive buffer
// check first if something received. If yes then set bit 15. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint16_t recv;

    servSystem();                           //serve DMX Reset, watchdog etc.

    if (gUSART_RXTAIL!=gUSART_RXHEAD)  {
        recv= (uint16_t)((*gUSART_RXTAIL++ & 0x00FF)|0x8000);
        if (gUSART_RXTAIL==gUSART_RXBUFTOP) {
                gUSART_RXTAIL=gUSART_RXBUF;
        }
        GPIOC->BSRR = GPIO_BSRR_BR_3;       // clear bit PC3, red LED OFF
        gRED_LEDCOUNT = 200;                //milliseconds
        return recv;
    }
    else return 0;
}


void sendCmdB(uint8_t data) // **********************************************
// **************************************************************************
{
    sendB2(data);
    if (guUsbCmdEnable) sendUsbByte(data);
}


void sendSetupNibble(uint8_t data) // ********************************************
// internally used to send 1 hex nibble
// **************************************************************************

{
uint8_t temp;

    temp = data+0x30;
    if (temp>0x39) temp+= 7;
    sendCmdB(temp);
}


void sendString(const uint8_t * str) // ********************************************
// ***************************************************************************
{
 uint8_t byte;

    while  ((byte = *str++) != 0)  sendCmdB(byte);
}


void sendX8(uint8_t data) // ************************************************
// send 1 byte in hex format (without leading 0x)
// **************************************************************************
{
    uint8_t temp;

    temp= (data&0xF0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendCmdB(temp);
    temp= data&0x0F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendCmdB(temp);
    sendCmdB(0x20);
}

void sendN8(uint8_t data) // ************************************************
// send 1 byte in hex format (without leading 0x), WITHOUT TRAILING SPACE
// **************************************************************************
{
    uint8_t temp;

    temp= (data&0xF0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendCmdB(temp);
    temp= data&0x0F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendCmdB(temp);
}

void sendX16(uint16_t data) // **********************************************
// send 1 half word (16bit) in hex format (without leading 0x)
// **************************************************************************
{
uint8_t nibble;

    nibble = (uint8_t) ((data>>12) & 0x000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) (data & 0x000F);
    sendSetupNibble(nibble);
    sendCmdB(0x20);
}


void sendX32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// **************************************************************************
{
uint8_t nibble;
    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
   sendSetupNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    sendSetupNibble(nibble);
    nibble = (uint8_t) (data & 0x0000000F);
    sendSetupNibble(nibble);
    sendCmdB(0x20);

}


void sendN32(uint32_t data) // **********************************************
//send signed decimal number string.
//Leading zeroes are suppressed, no trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
uint32_t quotient;

    if (data >= 0x80000000) {
        sendCmdB('-');
        data = ~data + 1 ;
    }
    if (data < 10) goto ONES;
    if (data < 100) goto TENS;
    if (data < 1000) goto HUNDREDS;
    if (data < 10000) goto THOUSANDS;
    if (data < 100000) goto TH10;
    if (data < 1000000) goto TH100;
    if (data < 10000000) goto MILLIONS;
    if (data < 100000000) goto MIL10;
    if (data < 1000000000) goto MIL100;
    quotient = data/1000000000;
    data = data - quotient*1000000000;
    sendCmdB((uint8_t)quotient+ 0x30);
MIL100:
    quotient = data/100000000;
    data = data - quotient*100000000;
    sendCmdB((uint8_t)quotient + 0x30);
MIL10:
    quotient = data/10000000;
    data = data - quotient*10000000;
    sendCmdB((uint8_t)quotient + 0x30);
MILLIONS:
    quotient = data/1000000;
    data = data - quotient*1000000;
    sendCmdB((uint8_t)quotient + 0x30);
TH100:
    quotient = data/100000;
    data = data - quotient*100000;
    sendCmdB((uint8_t)quotient+ 0x30);
TH10:
    quotient = data/10000;
    data = data - quotient*10000;
    sendCmdB((uint8_t)quotient + 0x30);
THOUSANDS:
    quotient = data/1000;
    data = data - quotient*1000;
    sendCmdB((uint8_t)quotient + 0x30);
HUNDREDS:
    quotient = data/100;
    data = data - quotient*100;
    sendCmdB((uint8_t)quotient+ 0x30);
TENS:
    quotient = data/10;
    data = data - quotient*10;
    sendCmdB((uint8_t)quotient + 0x30);
ONES:
    sendCmdB((uint8_t)data + 0x30) ;
}


void sendS32(uint32_t data) // **********************************************
//send signed decimal number string.
//Leading zeroes are suppressed, no trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendN32(data);
    sendCmdB(0x20);
}


uint16_t pullCmdRxBuf(void)  // *************************************
//receive byte from all channels in parallel with priority order USB > RS232
// ********************************************************************
{
volatile uint16_t recv;
	if (guConfigured)  {             //when active, USB has priority over USART
        if ((recv=pullUsbRxBuf()) != 0)  return recv;
    }
    if ((recv=pullUsartRxBuf()) != 0) return recv;
    return 0;
}


//##########################################################################

#ifdef DEBUG
// these routines can be used for non-blocking serial output of debug data

void sdbgB(uint8_t data) // *************************************************
// send 1 raw byte
// **************************************************************************
{
    if (gDBGHEAD < gDBGBUFTOP) *gDBGHEAD++ = data;
    if (gDBGHEAD == gDBGBUFTOP) {
        sendDebug();
        gDBGHEAD = gDBGBUF;

    }
}


void sdbgX8(uint8_t data) // ************************************************
// send 1 byte in hex format (without leading 0x)
// **************************************************************************
{
    uint8_t temp;

    temp= (data&0xF0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sdbgB(temp);
    temp= data&0x0F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sdbgB(temp);
    sdbgB(0x20);
}


void sdbgNibble(uint8_t data) // ********************************************
// internally used to send 1 hex nibble
// **************************************************************************

{
    uint8_t temp;

    temp = data+0x30;
    if (temp>0x39) temp+= 7;
    sdbgB(temp);
}


void sdbgX16(uint16_t data) // *********************************************
// send 1 half word (16bit) in hex format (without leading 0x)
// *************************************************************************
{
    uint8_t nibble;

    nibble = (uint8_t) ((data>>12) & 0x000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) (data & 0x000F);
    sdbgNibble(nibble);
    sdbgB(0x20);
}


void sdbgX32(uint32_t data) // *********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// *************************************************************************
{
    uint8_t nibble;

    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    sdbgNibble(nibble);
    nibble = (uint8_t) (data & 0x0000000F);
    sdbgNibble(nibble);
    sdbgB(0x20);

}


void sendDebug(void)  // ***************************************************
// *************************************************************************
{
    uint8_t * dbgTail = gDBGBUF;

    sendB(0x0D);
    sendB('D');
    sendB('E');
    sendB('B');
    sendB('U');
    sendB('G');
    sendB(':');
    sendB(0x20);
    while (dbgTail < gDBGHEAD) sendB(*dbgTail++);
    sendB('E');
    sendB('N');
    sendB('D');
    sendB(0x0D);
    gDBGHEAD = gDBGBUF;
}
#endif



//##########################################################################

// Interrupt Handlers
//*******************

void SysTick_Handler(void)  // **********************************************
// **************************************************************************
{
    if ((gSYSCOUNT > 0)&&(gSYSCOUNT < 0x8000)) gSYSCOUNT--; //slow servSystem()
    else gSYSCOUNT = 0;                     //double safety against underflow

    if (gGREEN_LEDCOUNT > 0) gGREEN_LEDCOUNT--;
    if (gRED_LEDCOUNT > 0) gRED_LEDCOUNT--;
    if (gDELAYCOUNT > 0) gDELAYCOUNT--;
    if (gTIMEOUT > 0) gTIMEOUT--;           //to be used in any subroutine

}


void USART2_IRQHandler(void)    // ******************************************
//for this special case, MIDI4_RXBUF is misused
// **************************************************************************
{
    if (USART2->ISR & USART_ISR_RXNE)  {
        *gUSART_RXHEAD++ = (uint8_t)(USART2->RDR & 0x00FF);
        if (gUSART_RXHEAD == gUSART_RXBUFTOP)  gUSART_RXHEAD=gUSART_RXBUF;
    }
}
